from copy import deepcopy
from timemachines.skatertools.data.ornstein import simulate_arima_like_path


def skater_bump(ys, f, num_points=51, ndx=-1, k=1):
    """ Sensitivity to last value  """
    # Pass e<0 for most of the data
    assert ndx<0
    import numpy as np
    y_bumped = [ ys[ndx] + 0.01 * i for i in np.linspace(-100,100,num_points)]

    # Roll forward to -ndx
    s = {}
    for y in ys[:ndx]:
        x, x_std, s = f(y=y, k=k, s=s, e=-1)

    # Process bumped observations
    x_final_values = list()
    for y_bump in y_bumped:
        s_ = deepcopy(s)
        # Process ndx, which might be the last
        e = 1000 if ndx==-1 else -1
        x, x_std, s_ = f(y=y_bump, k=k, s=s_, e=e)
        # Might need to process from ndx+1 to -2
        for y in ys[ndx:-2]:
            x, x_std, s_ = f(y=y, k=k, s=s_, e=-1)
        # Process the last with e=1000
        if ndx<-1:
            x, x_std, s_ = f(y=ys[-1], k=k, s=s_, e=1000)
        x_final_values.append(x[k-1])

    return y_bumped, x_final_values




# Compute performance rating for a single algorithm

from timemachines.skatertools.comparison.skaterelo import _init_elo
from timemachines.skatertools.recommendations.suggestions import get_ratings, show_ratings
from pprint import pprint
from timemachines.skatertools.comparison.skaterelo import skater_elo_multi_update, DEFAULT_INITIAL_ELO

DEFAULT_BENCHMARK_SKATER_NAMES = ['sluggish_moving_average',
                                  'thinking_fast_and_slow',
                                   'aggressive_ema_ensemble',
                                  'tsa_balanced_theta_ensemble',
                                  'thinking_fast_and_fast',
                                  'empirical_last_value',
                                  'rapidly_moving_average']


def performance_rating(skater_names:[str],k=1, category='univariate', initial_elo=DEFAULT_INITIAL_ELO,
                       benchmark_skater_names=None, update_frequency=10):
    # Play in a small pool that includes some rated skaters
    from timemachines.skaters.allskaters import skater_from_name
    if benchmark_skater_names is None:
        benchmark_skater_names = DEFAULT_BENCHMARK_SKATER_NAMES
    all_elo = get_ratings(k=k, category=category)
    ndx = [ i for i,nm in enumerate(all_elo['name']) if nm in benchmark_skater_names ]
    elo = dict([ (ky, [ v[i] for i in ndx] ) for ky,v in all_elo.items() if ky not in ['evaluator'] ])

    benchmark_skater_attempted = [ skater_from_name(nm) for nm in benchmark_skater_names ]
    benchmark_skater_population = [ f for f in benchmark_skater_attempted if not f is None ]
    benchmark_skater_failed = [ nm for nm, a in zip(benchmark_skater_names,benchmark_skater_attempted) if a is None]
    if len(benchmark_skater_failed):
        print('The following benchmark skaters failed to load. You might need to install some requisite package')
        pprint(benchmark_skater_failed)

    new_skater_attempted = [ skater_from_name(nm) for nm in skater_names ]
    new_skater_population = [ f for f in new_skater_attempted if not f is None ]
    new_skater_failed = [ nm for nm, a in zip(skater_names,new_skater_attempted) if a is None]
    if len(new_skater_failed):
        print('The following new skaters to load. You might need to install some requisite package')
        pprint(new_skater_failed)

    new_elo = _init_elo(elo={}, skater_population=new_skater_population, initial_elo=initial_elo)
    for ky in ['name','count','rating','traceback','active','pypi','seconds']:
        elo[ky] = elo[ky] + new_elo[ky]

    while True:
        import random
        if random.choice(range(update_frequency))==0:
            rat = show_ratings(elo)
            some_rat = [ r for r in rat if r[1] in benchmark_skater_names+skater_names]
            print('')
            print('---- Ratings -----')
            pprint(some_rat)
        elo = skater_elo_multi_update(elo=elo, k=k, evaluator=None, n_burn=400, tol=0.01, initial_elo=1600,
                                data_source=None, skater_population=benchmark_skater_population, always_skaters=None, verbose=False)


if __name__=='__main__':
    from timemachines.skaters.simple.thinking import WIGGLY_THINKING_FAST_AND_SLOW_SKATERS
    skater_names = [f.__name__ for f in WIGGLY_THINKING_FAST_AND_SLOW_SKATERS]
    performance_rating(skater_names=skater_names)



import numpy as np
from timemachines.skatertools.utilities.suppression import no_stdout_stderr
import random


def pull_towards_zero(xs, kappa=0.1):
    """ Take a sequence xs and apply mean reversion """
    dxs = np.diff(xs)
    xt = 0
    ou = [ xs[0] ]
    for dx in dxs:
        xt = xt + dx
        xt = xt - kappa*xt
        ou.append(xt)
    return np.array(ou)


def simulate_arima_like_path(seq_len, reverse=False):
    """  Just a way to generate a time-series that isn't completely trivial in structure
    :param seq_len:
    :param plot:
    :return:
    """
    import pandas as pd
    import statsmodels.api as sm

    # Load data
    y = sm.datasets.macrodata.load_pandas().data['cpi']
    for j in range(len(y)):
        y[j] += np.random.randn()

    y.index = pd.period_range('1959Q1', '2009Q3', freq='Q')

    # Create and fit a model once to generate data
    p = random.choice(range(2, 20))
    d = random.choice(range(0, 3))
    q = random.choice(range(0, 3))
    scale = random.choice([2, 1, 0.5, -0.5, -1, -2])
    with no_stdout_stderr():
        mod = sm.tsa.SARIMAX(y, order=(p, d, q), trend='c')
        res = mod.fit()
        res.model_orders['ar_true'] = p
        res.model_orders['ma_true'] = q
    from pprint import pprint
    pprint(res.model_orders)

    with no_stdout_stderr():
        sim = res.simulate(seq_len, anchor='end', repetitions=1)

    # Make sure the data doesn't go crazy in scale
    x = sim.reset_index()['cpi'].values.squeeze() * scale
    x = (x - x[-1]) / 10.0
    if reverse:
        ou = np.flip(pull_towards_zero(np.flip(x)))
    else:
        ou = pull_towards_zero(x)

    return ou


def show_example_ornstein():
    xs = np.cumsum(np.random.randn(500))
    y = pull_towards_zero(xs)
    import matplotlib.pyplot as plt
    t = list(range(len(y)))
    plt.plot(t,xs,t,y)
    plt.grid()
    plt.legend(['raw','ou'])
    plt.show()


def show_example_arima_like():
    y = simulate_arima_like_path(seq_len=500)
    import matplotlib.pyplot as plt
    t = list(range(len(y)))
    plt.plot(t,y)
    plt.grid()
    plt.legend(['raw','ou'])
    plt.show()


if __name__=='__main__':
    show_example_arima_like()







try:
    import merlion
    using_merlion = True
except ImportError:
    using_merlion = False
    
try:
    # Urgh... this doesn't succeed in determining that Merlion prophet will actually work.
    # Should perhaps instantiate a little prophet model to check
    import prophet
    from merlion.models.forecast.prophet import Prophet, ProphetConfig
    using_merlion_prophet = True
except ImportError:
    using_merlion_prophet = False

if __name__=='__main__':
    print(using_merlion)
    print(using_merlion_prophet)



from timemachines.skaters.sk.sfinclusion import using_statsforecast

if using_statsforecast:
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE
    from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory
    from timemachines.skaters.sk.sfautoarima import sf_autoarima


    def sf_autoarima_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        """ Chase residuals, somewhat cautiously using, quickly moving average """
        return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=sf_autoarima)



SF_HYPOCRATIC_SKATERS = [ sf_autoarima_hypocratic ]




try:
    import darts
    from darts.models import FFT, Prophet, ARIMA, AutoARIMA, ExponentialSmoothing, Theta, FourTheta, TransformerModel, \
        NBEATSModel
    using_darts = True
except (ImportError, OSError):
    using_darts = False




from timemachines.skaters.sk.sfinclusion import using_statsforecast

if using_statsforecast:
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE
    from timemachines.skaters.sk.sfautoarima import sf_autoarima
    from timemachines.skatertools.smoothing.wiggling import wiggler
    from timemachines.skatertools.combining.combiningforecasts import combine_using_mean,\
        combine_using_median, combine_using_huber

    def sf_autoarima_wiggly_factory(y: Y_TYPE, s: dict, k: int,
                                    a: A_TYPE = None, t: T_TYPE = None,
                                    e: E_TYPE = None, m=5, d=0.1, combiner=None):
        return wiggler(y=y, s=s, k=k, a=a, t=t, e=e, r=None, f=sf_autoarima, m=m, d=d, combiner=None)


    def sf_autoarima_wiggly_mean_d05_m5(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None,
                                        e: E_TYPE = None):
        return sf_autoarima_wiggly_factory(y=y, s=s, k=k, a=a, t=t, e=e, m=5, d=0.05, combiner=combine_using_mean)


    def sf_autoarima_wiggly_median_d05_m5(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None,
                                          e: E_TYPE = None):
        return sf_autoarima_wiggly_factory(y=y, s=s, k=k, a=a, t=t, e=e, m=5, d=0.05, combiner=combine_using_median)


    def sf_autoarima_wiggly_huber_d05_m5(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None,
                                         e: E_TYPE = None):
        return sf_autoarima_wiggly_factory(y=y, s=s, k=k, a=a, t=t, e=e, m=5, d=0.05, combiner=combine_using_huber)


    def sf_autoarima_wiggly_mean_d05_m3(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None,
                                         e: E_TYPE = None):
        return sf_autoarima_wiggly_factory(y=y, s=s, k=k, a=a, t=t, e=e, m=3, d=0.05, combiner=combine_using_mean)


    def sf_autoarima_wiggly_median_d05_m3(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None,
                                         e: E_TYPE = None):
        return sf_autoarima_wiggly_factory(y=y, s=s, k=k, a=a, t=t, e=e, m=3, d=0.05, combiner=combine_using_median)


    def sf_autoarima_wiggly_huber_d05_m3(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None,
                                         e: E_TYPE = None):
        return sf_autoarima_wiggly_factory(y=y, s=s, k=k, a=a, t=t, e=e, m=3, d=0.05, combiner=combine_using_huber)

    SF_AUTOARIMA_WIGGLY_SKATERS = [sf_autoarima_wiggly_mean_d05_m3,
                                   sf_autoarima_wiggly_median_d05_m3,
                                   sf_autoarima_wiggly_huber_d05_m3,
                                   sf_autoarima_wiggly_mean_d05_m5,
                                   sf_autoarima_wiggly_median_d05_m5,
                                   sf_autoarima_wiggly_huber_d05_m5] # make more if you need them

else:
    SF_AUTOARIMA_WIGGLY_SKATERS = []




try:
    from statsforecast.core import StatsForecast
    from sktime.forecasting.statsforecast import StatsForecastAutoARIMA
    using_statsforecast = True
except ImportError:
    using_statsforecast = False

if __name__=='__main__':
    print({'using_statsforecast':using_statsforecast})
    from sktime.forecasting.statsforecast import StatsForecastAutoARIMA